"""Codec for serialising and deserialising for Azure"""

import dataclasses
import json
from typing import Any, Dict

from llamazure.azgraph.models import Req, Res, ResErr, ResMaybe


class Encoder(json.JSONEncoder):
	"""Encode Req for JSON for Azure"""

	def default(self, o: Any) -> Any:
		if dataclasses.is_dataclass(o):
			return dataclasses.asdict(o)
		return super().default(o)


class Decoder:
	"""Decode Res from JSON from Azure"""

	def decode(self, req: Req, o: Dict) -> ResMaybe:
		"""Decode Res from JSON from Azure"""
		error = o.pop("error", None)
		if error:
			details_json = error.pop("details", [])
			details = tuple(detail_json for detail_json in details_json)
			return ResErr(**error, details=details)

		skip_token = o.pop("$skipToken", None)
		return Res(req=req, **o, skipToken=skip_token)



"""Interface to the Azure Resource Graph"""
from __future__ import annotations

import dataclasses
import json
import operator
from functools import reduce
from typing import Any, Tuple, Union, cast

import requests

from llamazure.azgraph import codec
from llamazure.azgraph.models import Req, Res, ResErr, ResMaybe


@dataclasses.dataclass
class RetryPolicy:
	"""Parameters and strategies for retrying Azure Resource Graph queries"""

	retries: int = 0  # number of times to retry. This is in addition to the initial try


class Graph:
	"""
	Access the Azure Resource Graph

	The easiest way to instantiate this is with the `from_credential` method.

	>>> from azure.identity import DefaultAzureCredential
	>>> credential = DefaultAzureCredential()
	>>> graph = Graph(credential)

	Making queries is easiest with the `q` method:
	>>> graph.q("Resources | project id, name, type, location | limit 5")
	[{'id': '/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/rg0/providers/Microsoft.Storage/storageAccounts/sa0', 'name': 'sa0', 'type': 'microsoft.storage/storageaccounts', 'location': 'canadacentral'}]

	If you want to provide options to the query, use a `Req` and the `query` function

	>>> from llamazure.azgraph.models import Req, Res
	>>> graph.query(Req(
	... 	query="Resources | project id, name, type, location | limit 5",
	... 	subscriptions=("00000000-0000-0000-0000-000000000001",)
	... ))
	"""

	def __init__(self, token, subscriptions: Tuple[str], retry_policy: RetryPolicy = RetryPolicy()):
		self.token = token
		self.subscriptions = subscriptions
		self.retry_policy = retry_policy

	@classmethod
	def from_credential(cls, credential) -> Graph:
		"""Create from an Azure credential"""
		token = credential.get_token("https://management.azure.com//.default")
		subscriptions = cls._get_subscriptions(token)
		return cls(token, subscriptions)

	@staticmethod
	def _get_subscriptions(token) -> Tuple[str]:
		raw = requests.get(
			"https://management.azure.com/subscriptions?api-version=2020-01-01",
			headers={"Authorization": f"Bearer {token.token}", "Content-Type": "application/json"},
		).json()
		return cast(Tuple[str], tuple(s["subscriptionId"] for s in raw["value"]))

	def q(self, q: str) -> Union[Any, ResErr]:
		"""Make a graph query"""
		res = self.query(Req(q, self.subscriptions))
		if isinstance(res, ResErr):
			return res
		return res.data

	def _exec_query(self, req) -> ResMaybe:
		raw = requests.post(
			"https://management.azure.com/providers/Microsoft.ResourceGraph/resources?api-version=2021-03-01",
			headers={"Authorization": f"Bearer {self.token.token}", "Content-Type": "application/json"},
			data=json.dumps(req, cls=codec.Encoder),
		).json()
		res = codec.Decoder().decode(req, raw)
		return res

	def query_single(self, req: Req) -> ResMaybe:
		"""Make a graph query for a single page"""
		res = self._exec_query(req)

		if isinstance(res, ResErr):
			retries = 0
			while retries < self.retry_policy.retries and isinstance(res, ResErr):
				retries += 1
				res = self._exec_query(req)
		return res

	def query_next(self, req: Req, previous: Res) -> ResMaybe:
		"""Query the next page in a paginated query"""
		options = req.options.copy()
		options["$skipToken"] = previous.skipToken

		# "$skip" overrides "$skipToken", so we need to remove it.
		# This is fine, since the original skip amount is encoded into the
		options.pop("$skip", None)

		next_req = dataclasses.replace(req, options=options)
		return self.query_single(next_req)

	def query(self, req: Req) -> ResMaybe:
		"""Make a graph query"""
		ress = []
		res = self.query_single(req)
		if isinstance(res, ResErr):
			return res

		ress.append(res)
		while res.skipToken:
			res = self.query_next(req, res)
			if isinstance(res, ResErr):
				return res
			ress.append(res)
		return reduce(operator.add, ress)



"""Models for the Azure Resource Graph"""
from __future__ import annotations

import dataclasses
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Tuple, TypedDict, Union


@dataclass(frozen=True)
class Req:
	"""Azure Resource Graph request"""

	query: str
	subscriptions: Tuple[str, ...]

	facets: Tuple = tuple()
	managementGroupId: Optional[str] = None
	options: Dict = field(default_factory=dict)


@dataclass(frozen=True)
class Res:
	"""Azure Resource Graph response"""

	req: Req

	totalRecords: int
	count: int
	resultTruncated: Any
	facets: Tuple
	data: Any
	skipToken: Optional[str] = None

	def __add__(self, other):
		if not isinstance(other, Res):
			raise TypeError(type(other))
		# using `other` here ensures that we get the skipToken and other stuff more up-to-date
		return dataclasses.replace(other, count=self.count + other.count, data=self.data + other.data)

	def to_optional(self) -> Optional[Any]:
		return self.data


@dataclass(frozen=True)
class ResErr:
	"""Azure Resource Graph error response"""

	code: str
	message: str
	details: Tuple[ErrorDetails, ...]

	def to_optional(self) -> Optional[Any]:
		return None


class ErrorDetails(TypedDict):
	"""Error details. TypedDict because extra parameters are supplied"""

	code: str
	message: str


ResMaybe = Union[Res, ResErr]




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: llamazure/azgraph:llamazure.azgraph

from setuptools import setup

setup(**{
    'author': 'Daniel Goldman',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Utilities',
        'Topic :: Internet :: Log Analysis',
    ],
    'description': 'Azure Resources Graph client',
    'install_requires': (
        'requests<3,>=2',
        'types-requests',
    ),
    'license': 'Round Robin 2.0.0',
    'long_description': """# llamazure.azgraph : Azure Resources Graph client

The `llamazure.azgraph` package provides a usable client for the Azure Resource Graph.

Benefits:
- automatically queries all your subscriptions
- no boilerplate
- easily navigate paginated queries

## azgraph

### Usage

Create a `Graph` with the `from_credential` and any of the standard Azure credentials.

```python
from azure.identity import DefaultAzureCredential

from llamazure.azgraph.azgraph import Graph

g = Graph.from_credential(DefaultAzureCredential())
```

#### Querying

Make a simple query with the `q` method, which will return your data directly:

```python
>>> g.q("Resources | project id, name, type, location | limit 5")
[{'id': '/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/rg0/providers/Microsoft.Storage/storageAccounts/sa0', 'name': 'sa0', 'type': 'microsoft.storage/storageaccounts', 'location': 'canadacentral'}]
```

Or specify options with the `query` method, which will return the full result object:

```python
>>> from llamazure.azgraph.models import Req

>>> g.query(Req(query="Resources | project id, name, type, location | limit 1", subscriptions=g.subscriptions, options={"$skip": 1},))
Res(req=Req(query='Resources | project id, name, type, location | limit 1', subscriptions=('00000000-0000-0000-0000-000000000000',), facets=(), managementGroupId=None, options={'$skip': 1}), totalRecords=1, count=1, resultTruncated='false', facets=[], data=[{'id': '/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/NetworkWatcherRG/providers/Microsoft.Network/networkWatchers/NetworkWatcher_canadacentral', 'name': 'NetworkWatcher_canadacentral', 'type': 'microsoft.network/networkwatchers', 'location': 'canadacentral'}], skipToken=None)
```

#### Retries

Every query can be automatically retried by the retry policy. You can modify this by setting the `Graph.retry_policy` attribute:

```python
g.retry_policy = RetryPolicy(retries=10)
```

#### Pagination

Pagination is handled automatically. If you want to manually paginate, you can manually walk the pages:

```python
req = Req(query="Resources | project id, name, type, location | limit 5", subscriptions=g.subscriptions)

res0 = g.query_single(req)
res1 = g.query_next(req, res0)
res2 = g.query_next(req, res1)
```
""",
    'long_description_content_type': 'text/markdown',
    'name': 'llamazure.azgraph',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'llamazure',
        'llamazure.azgraph',
    ),
    'python_requires': '>=3.8',
    'version': '0.0.3',
})
